@@ -9,6 +9,7 @@ if TYPE_CHECKING:
     from duckdb import Expression  # noqa: F401
 
     from narwhals._duckdb.dataframe import DuckDBLazyFrame  # noqa: F401
+    from narwhals._duckdb.expr import DuckDBWindowFunction
 
 
 class DuckDBSelectorNamespace(LazySelectorNamespace["DuckDBLazyFrame", "Expression"]):
@@ -20,12 +21,13 @@ class DuckDBSelectorNamespace(LazySelectorNamespace["DuckDBLazyFrame", "Expressi
 class DuckDBSelector(  # type: ignore[misc]
     CompliantSelector["DuckDBLazyFrame", "Expression"], DuckDBExpr
 ):
+    _window_function: DuckDBWindowFunction | None = None
+
     def _to_expr(self) -> DuckDBExpr:
         return DuckDBExpr(
             self._call,
             self._window_function,
             evaluate_output_names=self._evaluate_output_names,
             alias_output_names=self._alias_output_names,
-            backend_version=self._backend_version,
             version=self._version,
         )
