@@ -19,6 +19,7 @@ from narwhals._utils import (
     _into_arrow_table,
     convert_str_slice_to_int_slice,
     generate_temporary_column_name,
+    is_boolean_selector,
     is_compliant_series,
     is_index_selector,
     is_range,
@@ -276,9 +277,8 @@ class PolarsBaseFrame(Generic[NativePolarsFrame]):
         if order_by is None:
             result = frame.with_row_index(name)
         else:
-            end = pl.count() if self._backend_version < (0, 20, 5) else pl.len()
             result = frame.select(
-                pl.int_range(start=0, end=end).sort_by(order_by).alias(name), pl.all()
+                pl.int_range(pl.len()).over(order_by=order_by).alias(name), pl.all()
             )
 
         return self._with_native(result)
@@ -476,7 +476,11 @@ class PolarsDataFrame(PolarsBaseFrame[pl.DataFrame]):
             if not is_slice_none(columns):
                 if isinstance(columns, Sized) and len(columns) == 0:
                     return self.select()
-                if is_index_selector(columns):
+                if is_boolean_selector(columns):
+                    native = native.select(
+                        *(col for col, select in zip(native.columns, columns) if select)
+                    )
+                elif is_index_selector(columns):
                     if is_slice_index(columns) or is_range(columns):
                         native = native.select(
                             self.columns[slice(columns.start, columns.stop, columns.step)]
