@@ -23,7 +23,7 @@ class SparkLikeExprListNamespace(
 
     def contains(self, item: NonNestedLiteral) -> SparkLikeExpr:
         def func(expr: Column) -> Column:
-            F = self.compliant._F  # noqa: N806
+            F = self.compliant._F
             return F.array_contains(expr, F.lit(item))
 
         return self.compliant._with_elementwise(func)
